#lang racket/gui

;;; ----------------------------------------------------------------;;;
;;;                     A-IDE v0.2 source code                      ;;;
;;; Author: Antoine Bossard (Kanagawa University, Japan)            ;;;
;;; Licence: MIT licence                                            ;;;
;;; Homepage: https://www.sci.kanagawa-u.ac.jp/info/abossard/a-ide/ ;;;
;;;                    Copyright (c) A. Bossard                     ;;;
;;; ----------------------------------------------------------------;;;

; --- imports ---
(require framework)
(require parser-tools/lex)
(require "constants.rkt")

; --- exports ---
(provide code-text%
         )

; --- body ---
(define code-text%
  (class color:text%
    (super-new)
    (init frame
          name
          state
          line
          )
    (define main-frame frame)
    (define filename-label name)
    (define state-label state)
    (define line-label line)

    (define (token-sym->style sym)
      (cond [(eq? 'directive sym) "directive"]
            [(eq? 'instruction sym) "instruction"]
            [(eq? 'register sym) "register"]
            [(eq? 'numeric sym) "numeric"]
            [(eq? 'label sym) "label"]
            [(eq? 'comment sym) "comment"]
            [(eq? 'uncoloured sym) "base-font"]
            ))

    (define (get-token port offset mode)
      (define-tokens valued-tokens (directive instruction register numeric label comment))
      (define-empty-tokens empty-tokens (eof no-color uncoloured))
      (let* ([result
              ((lexer-src-pos
                [(eof) (token-eof)]

                [(concatenation (union ".equ" ".org" ".byte" ".ascii" ".string" ".include" ".text" ".section") whitespace) (token-directive lexeme)]
                [(union "hi8" "lo8") (token-directive lexeme)]
                [(concatenation
                  whitespace
                  (union
                   "add" "adc" "adiw" "sub" "subi" "sbc" "sbci" "sbiw"
                   "and" "andi" "or" "ori" "eor" "com" "neg"
                   "sbr" "cbr" "inc" "dec" "tst" "clr" "ser"
                   "mul" "muls" "mulsu" "fmul" "fmuls" "fmulsu" "des"
                   "rjmp" "ijmp" "eijmp" "jmp" "rcall" "icall" "eicall" "call" "ret" "reti"
                   "cpse" "cp" "cpc" "cpi" "sbrc" "sbrs" "sbic" "sbis"
                   "brbs" "brbc" "breq" "brne" "brcs" "brcc" "brsh" "brlo" "brmi" "brpl"
                   "brge" "brlt" "brhs" "brhc" "brts" "brtc" "brvs" "brvc" "brie" "brid"
                   "mov" "movw" "ldi" "lds" "ld" "ldd" "sts" "st" "std" "lpm" "elpm" "spm"
                   "in" "out" "push" "pop" "xch" "las" "lac" "lat"
                   "lsl" "lsr" "rol" "ror" "asr" "swap" "sbi" "cbi" "bst" "bld" "bset" "bclr"
                   "sec" "clc" "sen" "cln" "sez" "clz" "sei" "cli" "ses" "cls" "sev" "clv" "set" "clt" "seh" "clh"
                   "break" "nop" "sleep" "wdr") whitespace) (token-instruction lexeme)]
                [(concatenation #\r (repetition 1 2 numeric)) (token-register lexeme)]
                [(concatenation (repetition 1 +inf.0
                                            (intersection (complement #\;) graphic)
                                            ) #\:) (token-label lexeme)]

                [(concatenation #\" (complement (concatenation any-string #\" any-string)) #\") (token-uncoloured)]
                
                [(concatenation
                  (union
                   (repetition 1 +inf.0 numeric)
                   (concatenation "0x" (repetition 1 +inf.0 (union (char-range #\A #\F) (char-range #\a #\f) numeric)))
                   (concatenation "0b" (repetition 1 +inf.0 (union #\0 #\1)))
                   )) (token-numeric lexeme)]
                
                [(concatenation #\; (repetition 0 +inf.0 (char-complement #\newline))) (token-comment lexeme)]
                [(concatenation "/*" (complement (concatenation any-string "*/" any-string)) "*/") (token-comment lexeme)]
                
                [any-char (token-uncoloured)]
                )
               port)]
             [token (position-token-token result)]
             [start-pos-offset (position-offset (position-token-start-pos result))]
             [end-pos-offset (position-offset (position-token-end-pos result))]
             )
        (values (token-value token) (token-name token) #f start-pos-offset end-pos-offset 1 #f)
        ))

    (define/augment (on-edit-sequence)
      (send this reset-region 0 'end))
    (define/augment (after-edit-sequence)
      (send this reset-region 0 'end))

    (define/augment (after-insert start len)
      (send state-label set-when-unset))
    (define/augment (after-delete start len)
      (send state-label set-when-unset))
    
    (define/augment (after-load-file success?)
      (when success?
        (let ([path (send this get-filename)])
          (let-values ([(base name must-be-dir?) (split-path path)])
            (send filename-label set-label (path->string name))
            (send main-frame set-label (string-append application-name " - " (path->string path)))
            (send main-frame add-recent-entry (path->string name) (path->string path))
            ))
        (send state-label unset)))
    (define/augment (after-save-file success?)
      (after-load-file success?))

    (define (update-line-number)
      (let ([line-number (add1 (send this position-line (send this get-end-position)))])
        (send line-label set-when-needed line-number)))
    (define/override (on-default-char event)
      (super on-default-char event)
      (update-line-number))
    (define/override (on-default-event event)
      (super on-default-event event)
      (update-line-number))
    
    (send this set-max-undo-history 'forever)
    (send this set-tabs '() 4 #f)
    (send this set-paste-text-only #t)
    
    (let* ([style-list (send this get-style-list)]
           [basic-style (send style-list basic-style)]
           [font-delta (make-object style-delta%)]
           [font-style (begin (send font-delta set-family 'modern)
                              (send font-delta set-face "Consolas")
                              (send style-list find-or-create-style basic-style font-delta))]
           [directive-delta (make-object style-delta%)]
           [directive-style (begin (send directive-delta set-delta-foreground "DarkCyan")
                                   (send style-list find-or-create-style font-style directive-delta))]
           [instruction-delta (make-object style-delta% 'change-bold)]
           [instruction-style (begin (send instruction-delta set-delta-foreground "RoyalBlue")
                                     (send style-list find-or-create-style font-style instruction-delta))]
           [register-delta (make-object style-delta%)]
           [register-style (begin (send register-delta set-delta-foreground "Olive")
                                  (send style-list find-or-create-style font-style register-delta))]
           [numeric-delta (make-object style-delta%)]
           [numeric-style (begin (send numeric-delta set-delta-foreground "ForestGreen")
                                 (send style-list find-or-create-style font-style numeric-delta))]
           [label-delta (make-object style-delta%)]
           [label-style (begin (send label-delta set-delta-foreground "MediumVioletRed")
                               (send label-delta set-delta-background "LavenderBlush")
                               (send style-list find-or-create-style font-style label-delta))]
           [comment-delta (make-object style-delta%)]
           [comment-style (begin (send comment-delta set-delta-foreground "Chocolate")
                                 (send style-list find-or-create-style font-style comment-delta))]
           )
      (send this change-style font-style)

      (send style-list new-named-style "base-font" font-style)
      (send style-list new-named-style "directive" directive-style)
      (send style-list new-named-style "instruction" instruction-style)
      (send style-list new-named-style "register" register-style)
      (send style-list new-named-style "numeric" numeric-style)
      (send style-list new-named-style "label" label-style)
      (send style-list new-named-style "comment" comment-style)
      )
    
    (send this start-colorer token-sym->style get-token '())
    
    ))
