#lang racket/gui

;;; ----------------------------------------------------------------;;;
;;;                     A-IDE v0.2 source code                      ;;;
;;; Author: Antoine Bossard (Kanagawa University, Japan)            ;;;
;;; Licence: MIT licence                                            ;;;
;;; Homepage: https://www.sci.kanagawa-u.ac.jp/info/abossard/a-ide/ ;;;
;;;                    Copyright (c) A. Bossard                     ;;;
;;; ----------------------------------------------------------------;;;

; --- imports ---
(require framework)
(require "code-text.rkt")
(require "twostate-message.rkt")
(require "line-message.rkt")
(require "constants.rkt")

; --- exports ---
(provide ide-frame%
         )

; --- body ---
(define ide-frame%
  (class frame%
    (super-new [label application-name]
               [x (get-preference 'frame-x (lambda () frame-default-x) 'timestamp preferences-filename)]
               [y (get-preference 'frame-y (lambda () frame-default-y) 'timestamp preferences-filename)]
               [width (get-preference 'frame-width (lambda () frame-default-width) 'timestamp preferences-filename)]
               [height (get-preference 'frame-height (lambda () frame-default-height) 'timestamp preferences-filename)]
               )
    (define percentages (get-preference 'frame-percentages (lambda () frame-default-percentages) 'timestamp preferences-filename))
    (define recent-entries
      (let ([entries (get-preference 'recent-entries (lambda () '()) 'timestamp preferences-filename)])
        (filter (lambda (entry) (file-exists? (cdr entry))) entries)))
    
    (define/augment (can-close?)
      (let ([filename (send code-editor get-filename)])
        (if (or (and (not filename) (not (eq? (send code-editor get-text) "")))
                (and filename (send code-editor is-modified?)))
            (let ([result (message-box/custom "Warning" "There are unsaved modifications. Close anyway?"
                                              "Save and close" "Cancel" "Close without saving"
                                              this '(default=2 caution) 2)])
              (cond [(= result 1) (send code-editor save-file #f 'text)]
                    [(= result 2) #f]
                    [(= result 3) #t]))
            #t)))
    
    (define/augment (on-close)
      (when (and thread-id (not (thread-dead? thread-id)))
        (kill-thread thread-id))
      (if (send reset-item is-checked?)
          (put-preferences '(frame-x frame-y frame-width frame-height frame-percentages)
                           (list frame-default-x frame-default-y
                                 frame-default-width frame-default-height
                                 frame-default-percentages
                                 )
                           #f preferences-filename)
          (put-preferences '(frame-x frame-y frame-width frame-height frame-percentages)
                           (list (send this get-x) (send this get-y)
                                 (send this get-width) (send this get-height)
                                 (send v0-layout get-percentages)
                                 )
                           #f preferences-filename)
          )
      (put-preferences '(recent-entries linewrap avr-as avr-ld avr-objcopy avr-size avrdude avrdude-conf avr-as-isa-mcu avrdude-part avrdude-programmer avr-port)
                       (list recent-entries
                             (send linewrap-item is-checked?)
                             (send pref-avr-as get-value)
                             (send pref-avr-ld get-value)
                             (send pref-avr-objcopy get-value)
                             (send pref-avr-size get-value)
                             (send pref-avrdude get-value)
                             (send pref-avrdude-conf get-value)
                             (send pref-avr-as-isa-mcu get-selection)
                             (send pref-avrdude-part get-selection)
                             (send pref-avrdude-programmer get-selection)
                             (send pref-port get-value)
                             )
                       #f preferences-filename)
      )

    (define/override (on-drop-file path)
      (when (can-close?)
        (send code-editor load-file path 'text)))
    
    (define (populate-recent-menu)
      (for-each (lambda (item) (send item delete)) (send recent-menu get-items))
      (if (null? recent-entries)
          (send recent-menu enable #f)
          (begin
            (for-each (lambda (entry)
                        (let ([filename (car entry)] [path (cdr entry)])
                          (new menu-item% [label filename] [parent recent-menu]
                               [callback (lambda (menu-item event) (when (and (file-exists? path) (can-close?)) (send code-editor load-file path 'text)))])))
                      recent-entries)
            (send recent-menu enable #t))))
    
    (define (add-recent-entry filename path)
      (let ([entry (cons filename path)])
        (when (not (member entry recent-entries))
          (set! recent-entries (cons entry (if (> (length recent-entries) 4) (take recent-entries 4) recent-entries)))
          (populate-recent-menu))))
    (public add-recent-entry)

    (define (callback-find str dir case?)
      (let ([position (if (non-empty-string? str) (send code-editor find-string str dir 'start 'eof #t case?) #f)])
        (when position
          (if (eq? dir 'forward)
              (send code-editor set-position position (+ position (string-length str)))
              (send code-editor set-position (- position (string-length str)) position)))))

    (define (callback-replace str-from str-to dir case?)
      (let ([position (if (non-empty-string? str-from) (send code-editor find-string str-from dir 'start 'eof #t case?) #f)])
        (when position
          (if (eq? dir 'forward)
              (begin
                (send code-editor delete position (+ position (string-length str-from)))
                (send code-editor insert str-to position)
                (send code-editor set-position position (+ position (string-length str-to)))
                )
              (let ([left-pos (- position (string-length str-from))])
                (send code-editor delete left-pos position)
                (send code-editor insert str-to left-pos)
                (send code-editor set-position left-pos (+ left-pos (string-length str-to)))
                )
              ))))
    
    (define (callback-toggle)
      (if (= (length (send v0-layout get-children)) 2)
          (begin
            (set! percentages (send v0-layout get-percentages))
            (send v0-layout delete-child output-canvas))
          (begin
            (send v0-layout add-child output-canvas)
            (send v0-layout set-percentages percentages))))

    (define (callback-homepage)
      (let ([url "https://www.sci.kanagawa-u.ac.jp/info/abossard/a-ide/"] [os (system-type)])
        (let*-values ([(command args) (cond [(eq? os 'unix) (values "xdg-open" (list url))]
                                            [(eq? os 'macosx) (values "open" (list url))]
                                            [(eq? os 'windows) (values "cmd.exe" (list "/c" "start" url))])]
                      [(process-id port-stdout port-stdin port-stderr) (apply subprocess #f #f #f (find-executable-path command) args)])
          (close-input-port port-stdout)
          (close-output-port port-stdin)
          (close-input-port port-stderr)
          )))
    
    (define (callback-about)
      (message-box (string-append "About " application-name)
                   "A simple IDE for AVR programming in assembly.\n\nA-IDE is distributed under the terms of the MIT licence.\nAVR is a registered trademark of Microchip Technology.\n\nVersion: 0.2\n©2023 A. Bossard"
                   this))
    (public callback-about)

    (define thread-id #f)
    (define (callback-build upload-after-build?)
      (let* ([filename (send code-editor get-filename)]
             [do-build
              (if (not filename)
                  (let ([result (message-box/custom "Warning" "Building requires saving to a file. Continue?"
                                                    "Save and build" "Cancel" #f this '(default=1 caution) 2)])
                    (cond [(= result 1) (send code-editor save-file #f 'text)]
                          [(= result 2) #f]))
                  (if (send code-editor is-modified?)
                      (let ([result (message-box/custom "Warning" "There are unsaved modifications. Build anyway?"
                                                        "Save and build" "Cancel" "Build without saving" this '(default=1 caution) 2)])
                        (cond [(= result 1) (send code-editor save-file #f 'text)]
                              [(= result 2) #f]
                              [(= result 3) #t]))
                      #t))])
        (when (and do-build (or (not thread-id) (thread-dead? thread-id)))
          (set! thread-id
                (thread (lambda ()
                          (let ([to-disable (list build-button upload-button build-item upload-item)]
                                [path (send code-editor get-filename)])
                            (for-each (lambda (control) (send control enable #f)) to-disable)
                            (begin-busy-cursor)
                            (send build-message set-label "building...")
                            (send output-editor lock #f)
                            (send output-editor erase)
                            (send output-editor change-style output-editor-style)
                            (send output-editor insert "--- building has started ---\n")
                            
                            (let* ([path-s path]
                                   [path-o (path-replace-extension path ".o")]
                                   [path-elf (path-replace-extension path ".elf")]
                                   [path-hex (path-replace-extension path ".hex")]
                                   [as-path (find-executable-path (send pref-avr-as get-value))]
                                   [ld-path (find-executable-path (send pref-avr-ld get-value))]
                                   [objcopy-path (find-executable-path (send pref-avr-objcopy get-value))]
                                   [size-path (find-executable-path (send pref-avr-size get-value))])
                              (if (and as-path ld-path objcopy-path size-path)
                                  (let ([commands
                                         (list
                                          (cons 'as (cons as-path (avr-as-parameters (send pref-avr-as-isa-mcu get-string-selection) path-o path-s)))
                                          (cons 'ld (cons ld-path (avr-ld-parameters path-elf path-o)))
                                          (cons 'objcopy (cons objcopy-path (avr-objcopy-parameters path-elf path-hex)))
                                          (cons 'size (cons size-path (avr-size-parameters path-elf)))
                                          )])
                                    (for-each (lambda (command)
                                                (let ([path-string (cadr command)] [arg-list (cddr command)])
                                                  (let-values ([(process-id port-stdout port-stdin port-stderr)
                                                                (apply subprocess #f #f #f path-string arg-list)])
                                                    (if (eq? (car command) 'size)
                                                        (begin
                                                          (subprocess-wait process-id)
                                                          (let ([size-result (string-trim (port->string port-stdout))])
                                                            (if (non-empty-string? size-result)
                                                                (let* ([total-line (last (string-split size-result "\n"))]
                                                                       [bytes (string->number (last (string-split total-line)))]
                                                                       [bytes-str (if (< bytes 1000) (format "Machine code: ~a bytes" bytes)
                                                                                      (format "Machine code: ~a Kbytes" (/ bytes 1000)))])
                                                                  (send build-message set-label bytes-str))
                                                                (send build-message set-label "")
                                                                ))
                                                          (close-output-port port-stdin)
                                                          (close-input-port port-stderr)
                                                          )
                                                        (do ([port-outerr (merge-input port-stdout port-stderr)])
                                                          ((not (eq? (subprocess-status process-id) 'running))
                                                           (close-input-port port-outerr)
                                                           (close-input-port port-stdout)
                                                           (close-output-port port-stdin)
                                                           (close-input-port port-stderr))
                                                          (do ([line-outerr (read-line port-outerr 'any) (read-line port-outerr 'any)]
                                                               )
                                                            ((eof-object? line-outerr)
                                                             )
                                                            (send output-editor insert line-outerr)
                                                            (send output-editor insert #\newline)
                                                            ))
                                                        ))))
                                              commands)
                                    )
                                  (send output-editor insert " ! Building has failed: some paths of the toolchain are incorrect.\n")
                                  )
                              
                              (send output-editor insert "\n--- building has completed ---\n")

                              (when upload-after-build?
                                (send output-editor insert "--- uploading has started ---\n")
                                (let* ([path-string (find-executable-path (send pref-avrdude get-value))]
                                       [avrdude-conf-path (find-executable-path (send pref-avrdude-conf get-value))]
                                       [arg-list (avrdude-parameters avrdude-conf-path (car (list-ref avrdude-parts (send pref-avrdude-part get-selection)))
                                                                     (send pref-avrdude-programmer get-string-selection) (send pref-port get-value) path-hex)])
                                  (if (and path-string avrdude-conf-path)
                                      (let-values ([(process-id port-stdout port-stdin port-stderr)
                                                    (apply subprocess #f #f #f path-string arg-list)])
                                        (set! avrdude-process-id process-id)
                                        (send kill-item enable #t)
                                        (do ([port-outerr (merge-input port-stdout port-stderr)])
                                          ((not (eq? (subprocess-status process-id) 'running))
                                           (close-input-port port-outerr)
                                           (close-input-port port-stdout)
                                           (close-output-port port-stdin)
                                           (close-input-port port-stderr))
                                          (do ([line-outerr (read-line port-outerr 'any) (read-line port-outerr 'any)])
                                            ((eof-object? line-outerr)
                                             )
                                            (send output-editor insert line-outerr)
                                            (send output-editor insert #\newline)
                                            ))
                                        (send kill-item enable #f)
                                        (set! avrdude-process-id #f)
                                        )
                                      (send output-editor insert " ! Uploading has failed: the path of avrdude or avrdude.conf is incorrect.\n")
                                      )
                                  )
                                (send output-editor insert "\n--- uploading has completed ---\n"))
                              )
                            
                            (send output-editor lock #t)
                            (end-busy-cursor)
                            (for-each (lambda (control) (send control enable #t)) to-disable)
                            )
                          ))))
        ))

    (define avrdude-process-id #f)
    (define (callback-kill)
      (when avrdude-process-id
        (subprocess-kill avrdude-process-id #t)
        (set! avrdude-process-id #f)))
    
    (define search-frame (new frame% [label "Find and replace"] [parent this]
                              [min-width 250] [border 2] [style '(no-resize-border)]
                              [stretchable-width #f] [stretchable-height #f]))
    (define search-edit (new text-field% [label "Search:"] [parent search-frame]))
    (define replace-edit (new text-field% [label "Replace with:"] [parent search-frame]))
    (define search-group (new group-box-panel% [label "Parameters"] [parent search-frame] [alignment '(left top)]))
    (define direction-radio (new radio-box% [label "Direction:"] [parent search-group] [choices '("Forward" "Backward")] [style '(horizontal)]))
    (define case-checkbox (new check-box% [label "Case-sensitive"] [parent search-group]))
    (define search-horizontal1 (new horizontal-pane% [parent search-frame] [alignment '(center center)]))
    (define search-button (new button% [label "Find"] [parent search-horizontal1] [style '(border)]
                               [callback (lambda (button event) (send search-frame show #f)
                                           (callback-find (send search-edit get-value)
                                                          (if (= (send direction-radio get-selection) 0) 'forward 'backward)
                                                          (send case-checkbox get-value)))]))
    (define replace-button (new button% [label "Replace"] [parent search-horizontal1]
                                [callback (lambda (button event) (send search-frame show #f)
                                            (callback-replace (send search-edit get-value)
                                                              (send replace-edit get-value)
                                                              (if (= (send direction-radio get-selection) 0) 'forward 'backward)
                                                              (send case-checkbox get-value)))]))
    
    (define (validate-path edit)
      (let ([str (send edit get-value)])
        (if (and (non-empty-string? str) (find-executable-path str))
            (send edit set-field-background (send the-color-database find-color "LightGreen"))
            (send edit set-field-background (send the-color-database find-color "LightPink")))))
    (define pref-frame (new frame% [label "Preferences"] [parent this] [width 400] [border 2] [stretchable-height #f]))
    (define pref-group (new group-box-panel% [label "Paths"] [parent pref-frame] [alignment '(left top)]))
    (define pref-horizontal1 (new horizontal-pane% [parent pref-group]))
    (define pref-avr-as (new text-field% [label "avr-as:"] [parent pref-horizontal1]
                             [callback (lambda (edit event) (when (eq? (send event get-event-type) 'text-field) (validate-path edit)))]))
    (new button% [label "..."] [parent pref-horizontal1]
         [callback (lambda (button event) (let ([path (get-file "Select avr-as" pref-frame)]) (when path (send pref-avr-as set-value (path->string path)) (validate-path pref-avr-as))))])
    (define pref-horizontal2 (new horizontal-pane% [parent pref-group]))
    (define pref-avr-ld (new text-field% [label "avr-ld:"] [parent pref-horizontal2]
                             [callback (lambda (edit event) (when (eq? (send event get-event-type) 'text-field) (validate-path edit)))]))
    (new button% [label "..."] [parent pref-horizontal2]
         [callback (lambda (button event) (let ([path (get-file "Select avr-ld" pref-frame)]) (when path (send pref-avr-ld set-value (path->string path)) (validate-path pref-avr-ld))))])
    (define pref-horizontal3 (new horizontal-pane% [parent pref-group]))
    (define pref-avr-objcopy (new text-field% [label "avr-objcopy:"] [parent pref-horizontal3]
                                  [callback (lambda (edit event) (when (eq? (send event get-event-type) 'text-field) (validate-path edit)))]))
    (new button% [label "..."] [parent pref-horizontal3]
         [callback (lambda (button event) (let ([path (get-file "Select avr-objcopy" pref-frame)]) (when path (send pref-avr-objcopy set-value (path->string path)) (validate-path pref-avr-objcopy))))])
    (define pref-horizontal4 (new horizontal-pane% [parent pref-group]))
    (define pref-avr-size (new text-field% [label "avr-size:"] [parent pref-horizontal4]
                               [callback (lambda (edit event) (when (eq? (send event get-event-type) 'text-field) (validate-path edit)))]))
    (new button% [label "..."] [parent pref-horizontal4]
         [callback (lambda (button event) (let ([path (get-file "Select avr-size" pref-frame)]) (when path (send pref-avr-size set-value (path->string path)) (validate-path pref-avr-size))))])
    (define pref-horizontal5 (new horizontal-pane% [parent pref-group]))
    (define pref-avrdude (new text-field% [label "avrdude:"] [parent pref-horizontal5]
                              [callback (lambda (edit event) (when (eq? (send event get-event-type) 'text-field) (validate-path edit)))]))
    (new button% [label "..."] [parent pref-horizontal5]
         [callback (lambda (button event) (let ([path (get-file "Select avrdude" pref-frame)]) (when path (send pref-avrdude set-value (path->string path)) (validate-path pref-avrdude))))])
    (define pref-horizontal6 (new horizontal-pane% [parent pref-group]))
    (define pref-avrdude-conf (new text-field% [label "avrdude.conf:"] [parent pref-horizontal6]
                                   [callback (lambda (edit event) (when (eq? (send event get-event-type) 'text-field) (validate-path edit)))]))
    (new button% [label "..."] [parent pref-horizontal6]
         [callback (lambda (button event) (let ([path (get-file "Select avrdude.conf" pref-frame)]) (when path (send pref-avrdude-conf set-value (path->string path)) (validate-path pref-avrdude-conf))))])
    (define parameters-group (new group-box-panel% [label "Parameters"] [parent pref-frame] [alignment '(left top)]))
    (define pref-avr-as-isa-mcu (new choice% [label "avr-as ISA/MCU: "] [parent parameters-group] [choices avr-as-isa-mcu] [stretchable-width #t]))
    (define pref-avrdude-part (new choice% [label "avrdude MCU: "] [parent parameters-group] [choices (map cdr avrdude-parts)] [stretchable-width #t]))
    (define pref-avrdude-programmer (new choice% [label "programmer: "] [parent parameters-group] [choices avrdude-programmer] [stretchable-width #t]))
    (define pref-port (new text-field% [label "Port:"] [parent parameters-group]))
    (define pref-button (new button% [label "OK"] [parent pref-frame] [callback (lambda (button event) (send pref-frame show #f))]))
    
    (define main-menu (new menu-bar% [parent this]))
    (define file-menu (new menu% [label "&File"] [parent main-menu]))
    (new menu-item% [label "&Open..."] [parent file-menu] [shortcut #\o] [callback (lambda (menu-item event) (when (can-close?) (send code-editor load-file "" 'text)))])
    (define recent-menu (new menu% [label "Open recen&t"] [parent file-menu]))
    (new menu-item% [label "&Save"] [parent file-menu] [shortcut #\s] [callback (lambda (menu-item event) (send code-editor save-file #f 'text))])
    (new menu-item% [label "Save &as..."] [parent file-menu] [callback (lambda (menu-item event) (send code-editor save-file "" 'text))])
    (new menu-item% [label "&Reload"] [parent file-menu] [callback (lambda (menu-item event) (when (can-close?) (send code-editor load-file #f 'text)))])
    (new separator-menu-item% [parent file-menu])
    (new menu-item% [label "&Preferences..."] [parent file-menu] [callback (lambda (menu-item event) (send pref-frame center) (send pref-frame show #t))])
    (new separator-menu-item% [parent file-menu])
    (new menu-item% [label "&Exit"] [parent file-menu] [shortcut #\q] [callback (lambda (menu-item event) (send this on-exit))])
    (define edit-menu (new menu% [label "&Edit"] [parent main-menu]))
    (append-editor-operation-menu-items edit-menu)
    (new separator-menu-item% [parent edit-menu])
    (new menu-item% [label "&Find and replace..."] [parent edit-menu] [shortcut #\f] [callback (lambda (menu-item event) (send search-frame center) (send search-edit focus) (send search-frame show #t))])
    (new menu-item% [label "Find &next"] [parent edit-menu] [shortcut #\g] [callback (lambda (menu-item event) (callback-find (send search-edit get-value) 'forward (send case-checkbox get-value)))])
    (new menu-item% [label "Find &previous"] [parent edit-menu] [shortcut-prefix '(ctl shift)] [shortcut #\g] [callback (lambda (menu-item event) (callback-find (send search-edit get-value) 'backward (send case-checkbox get-value)))])
    (new menu-item% [label "Replace next"] [parent edit-menu] [shortcut #\r] [callback (lambda (menu-item event) (callback-replace (send search-edit get-value) (send replace-edit get-value) 'forward (send case-checkbox get-value)))])
    (new menu-item% [label "Replace previous"] [parent edit-menu] [shortcut-prefix '(ctl shift)] [shortcut #\r] [callback (lambda (menu-item event) (callback-replace (send search-edit get-value) (send replace-edit get-value) 'backward (send case-checkbox get-value)))])
    (define view-menu (new menu% [label "&View"] [parent main-menu]))
    (define linewrap-item (new checkable-menu-item% [label "Line wrap"] [parent view-menu]
                               [callback (lambda (menu-item event) (send code-editor auto-wrap (send menu-item is-checked?)))]
                               [checked (get-preference 'linewrap (lambda () default-linewrap) 'timestamp preferences-filename)]))
    (new menu-item% [label "&Toggle output"] [parent view-menu] [shortcut #\e] [callback (lambda (menu-item event) (callback-toggle))])
    (new separator-menu-item% [parent view-menu])
    (define reset-item (new checkable-menu-item% [label "&Reset on restart"] [parent view-menu] [callback (lambda (menu-item event) (void))]))
    (define tools-menu (new menu% [label "&Tools"] [parent main-menu]))
    (define build-item (new menu-item% [label "&Build"] [parent tools-menu] [shortcut #\b] [callback (lambda (menu-item event) (send build-button command event))]))
    (define upload-item (new menu-item% [label "&Upload"] [parent tools-menu] [shortcut #\u] [callback (lambda (menu-item event) (send upload-button command event))]))
    (define kill-item (new menu-item% [label "&Stop upload"] [parent tools-menu] [shortcut-prefix '(ctl shift)] [shortcut #\u] [callback (lambda (menu-item event) (callback-kill))]))
    (define help-menu (new menu% [label "&Help"] [parent main-menu]))
    (new menu-item% [label "&Homepage"] [parent help-menu] [callback (lambda (menu-item event) (callback-homepage))])
    (new menu-item% [label "&About"] [parent help-menu] [shortcut 'f1] [shortcut-prefix '()] [callback (lambda (menu-item event) (callback-about))])

    (define v0-layout (new panel:vertical-dragable% [parent this]))
    (define v1-layout (new vertical-pane% [parent v0-layout]))
    (define h0-layout (new horizontal-pane% [parent v1-layout] [stretchable-height #f]))
    (define line-label (new line-message% [parent h0-layout] [horiz-margin 10]))
    (define filename-label (new message% [label "(new buffer)"] [parent h0-layout] [horiz-margin 10] [auto-resize #t]))
    (define state-label (new twostate-message% [state #f] [parent h0-layout] [auto-resize #t]))
    
    (define h1-layout (new horizontal-pane% [parent h0-layout] [alignment '(center center)]))
    (define build-message (new message% [label ""] [parent h1-layout] [horiz-margin 10] [auto-resize #t]))
    (define h2-layout (new horizontal-pane% [parent h0-layout] [alignment '(right center)]))
    (define build-button (new button% [label "&Build"] [parent h2-layout] [callback (lambda (button event) (callback-build #f))]))
    (define upload-button (new button% [label "&Upload"] [parent h2-layout] [callback (lambda (button event) (callback-build #t))]))

    (define code-editor (new code-text% [frame this] [name filename-label] [state state-label] [line line-label]))
    (define main-canvas (new editor-canvas% [parent v1-layout] [editor code-editor]  [style '(auto-hscroll)]))

    (define output-editor (new text%))
    (define output-canvas (new editor-canvas% [parent v0-layout] [editor output-editor] [style '(auto-hscroll)]))

    (define editor-menu (new popup-menu%))

    (define common-keymap (new keymap%))
    (keymap:setup-editor common-keymap)
    (add-text-keymap-functions common-keymap)
    (send common-keymap map-function "home" "beginning-of-line")
    (send common-keymap map-function "end" "end-of-line")
    (send common-keymap map-function "s:home" "select-to-beginning-of-line")
    (send common-keymap map-function "s:end" "select-to-end-of-line")
    (send common-keymap map-function "c:right" "forward-word")
    (send common-keymap map-function "c:left" "backward-word")
    (send common-keymap map-function "c:s:right" "forward-select-word")
    (send common-keymap map-function "c:s:left" "backward-select-word")
    (send common-keymap map-function "c:k" "delete-to-end-of-line")
    (send common-keymap map-function "c:delete" "delete-line")
    
    (send common-keymap add-function "show-popup"
          (lambda (editor event)
            (when (send event button-up? 'right)
              (for-each (lambda (item) (send item delete)) (send editor-menu get-items))
              (append-editor-operation-menu-items editor-menu
                                                  (list editor (send editor get-start-position)))
              (let-values ([(edt-x edt-y) (send editor dc-location-to-editor-location (send event get-x) (send event get-y))])
                (send (send editor get-admin) popup-menu editor-menu edt-x edt-y))
              )))
    
    (send common-keymap map-function "rightbuttonseq" "show-popup")

    (if (= (length percentages) 1)
        (callback-toggle)
        (send v0-layout set-percentages percentages))

    (send kill-item enable #f)
    (populate-recent-menu)
    
    (send code-editor set-keymap common-keymap)
    (when (send linewrap-item is-checked?) (send code-editor auto-wrap #t))

    (send output-canvas set-canvas-background (send the-color-database find-color "Gainsboro"))

    (define output-editor-style
      (let* ([style-list (send output-editor get-style-list)]
             [original-style (send (send code-editor get-style-list) find-named-style "base-font")])
        (send style-list convert original-style)))
    (send output-editor change-style output-editor-style)
    (send output-editor set-keymap common-keymap)
    (send output-editor auto-wrap #t)
    (send output-editor lock #t)

    (send pref-avr-as set-value (get-preference 'avr-as (lambda () default-avr-as) 'timestamp preferences-filename))
    (validate-path pref-avr-as)
    (send pref-avr-ld set-value (get-preference 'avr-ld (lambda () default-avr-ld) 'timestamp preferences-filename))
    (validate-path pref-avr-ld)
    (send pref-avr-objcopy set-value (get-preference 'avr-objcopy (lambda () default-avr-objcopy) 'timestamp preferences-filename))
    (validate-path pref-avr-objcopy)
    (send pref-avr-size set-value (get-preference 'avr-size (lambda () default-avr-size) 'timestamp preferences-filename))
    (validate-path pref-avr-size)
    (send pref-avrdude set-value (get-preference 'avrdude (lambda () default-avrdude) 'timestamp preferences-filename))
    (validate-path pref-avrdude)
    (send pref-avrdude-conf set-value (get-preference 'avrdude-conf (lambda () default-avrdude-conf) 'timestamp preferences-filename))
    (validate-path pref-avrdude-conf)
    (send pref-avr-as-isa-mcu set-selection (get-preference 'avr-as-isa-mcu (lambda () default-avr-as-isa-mcu) 'timestamp preferences-filename))
    (send pref-avrdude-part set-selection (get-preference 'avrdude-part (lambda () default-avrdude-part) 'timestamp preferences-filename))
    (send pref-avrdude-programmer set-selection (get-preference 'avrdude-programmer (lambda () default-avrdude-programmer) 'timestamp preferences-filename))
    (send pref-port set-value (get-preference 'avr-port (lambda () default-port) 'timestamp preferences-filename))
    
    (send this accept-drop-files #t)
    (send main-canvas focus)
    ))
